; $Workfile: Dsp54Laws.asm $   $Revision: 2 $    $Modtime: 5/28/03 4:38p $
;*------------------------------------------------------- File Description -*\
;
;   Odin TeleSystems Inc.
;
;   SAA-1007-1 OTX DSP C54x Software Development Kit (SDK)
;
;   Law (A and u) Companding ASM subroutines called from C functions.
;
;   For NEAR mode compile with:
;
;     asm500 -dFARMODE=0 Dsp54Laws.asm
;
;   For FAR mode compile with:
;
;     asm500 -dFARMODE=1 -mf -v548 Dsp54Laws.asm
;
;*--------------------------------------------------------------------------*/

        .mmregs

        .global _OtxDspAlawCompress
        .global _OtxDspUlawCompress
        .global _OtxDspLawExpand

		.if FARMODE==1
RETOFS  .set	1				; in far mode the return address pushed on the stack is two addresses, therefore the function parameters (which are pushed on the stack) are offset by 1  
		.asg "fret", RETCMD
		.else
RETOFS  .set	0
		.asg "ret", RETCMD
		.endif


ULAWBIAS   .set    33           ;u-law compression bias value
ULAWMASK   .set    7fh          ;PCM u-law transmission mask (invert all bits)
ALAWMASK   .set    55h          ;PCM A-law transmission mask (invert every other bit)
PCMSIGN    .set    80h          ;PCM polarity bit -> B7=0 (pos), B7=1 (neg)


        .text



*--------------------------------------------------------------------------*
*   u/a-LAW Expanding
*
*
*  Input:
*
*    lawExpandInput (8-bits sign magnitude value)
*    law (0=ulaw, 1=alaw)
*
*  Returns:
*
*    14-bit 2-s complement value (in the 14 least significant bits of a
*                                 16-bit word)
*
*  Note:
*
*    Since a-law only has a range of -4196 < x 4195, we multiply the value
*    to be returned by 2 before it is returned to make it a full scale
*    14-bit value.
*
*--------------------------------------------------------------------------*
_OtxDspLawExpand
        nop
        pshm    ST0             ; *SP(6) = ST0
        pshm    ST1             ; *SP(5) = ST1
        pshm    AR1             ; *SP(4) = AR1
        pshm    AR6             ; *SP(3) = AR6
        pshm    AR7             ; *SP(2) = AR7
        nop
        frame   #-2             ;---- local variables -----
                                ; *SP(1) = not used
                                ; *SP(0) = temp variable
                                ;---- arguments -----------
                                ; (A)    = lawExpandInput: 8-bit value to be decompanded to 16-bit
                                ; *SP(8+RETOFS) = law: 0=ulaw, 1=alaw
                                ; *SP(7) = return address

        ssbx    SXM             ; added by HJ/NDF 971219 to eliminate optimization problem

        and     #00FFh,A        ; We should have an 8-bit value in A now
        bitf    *SP(8+RETOFS),1h       ; check which law (0=ulaw, 1=alaw)
        nop
        xc      2,TC            ; if law=0 then use table at 0xfc00 (ulaw)
                                ; since <add #lk,dst> in next line is 2 words, n=2 for xc
         add     #256,A         ; if law=1 then offset by 256 --> use table at 0xfd00 (alaw)


        add     #0fc00h,A       ; Add offset to find correct place in table
        and     #0FFFFh, A      ; Clear upper bits in A
        reada   *SP(0)          ; Transfer lookup table value (prog mem) to data memory
        ld      *SP(0),A        ; Put the look up value in A
        bitf    *SP(8),1h       ; check which law (0=ulaw, 1=alaw)
        nop
        xc      1,TC            ; if law=1 (a-law) then he is a 13 bit value. Need to make him a 14 bit value.
         sfta   A, 1            ; Multiply by 2 to make a-law value a 14-bit value.


        frame   #2
        nop
        popm    AR7
        popm    AR6
        popm    AR1
        popm    ST1
        popm    ST0
        RETCMD



*--------------------------------------------------------------------------*
*   u-LAW Compression
*
*    Input:
*
*      lawCompInput (14 bit 2-s complement value)
*
*    Returns:
*
*      u-law compressed value ready to transmission (8-bits)
*
*    Note:
*
*      Value bits (0-6) are inverted.
*
*--------------------------------------------------------------------------*
_OtxDspUlawCompress
        nop
        pshm    ST0             ; *SP(6) = ST0
        pshm    ST1             ; *SP(5) = ST1
        pshm    AR1             ; *SP(4) = AR1
        pshm    AR6             ; *SP(3) = AR6
        pshm    AR7             ; *SP(2) = AR7
        nop
        frame   #-2             ;---- local variables -----
                                ; *SP(1) = exponent
                                ; *SP(0) = original lawCompInput saved here
                                ;---- arguments -----------
                                ; (A)    = lawCompInput: 16-bit value to be companded
                                ; *SP(7) = return address
        ssbx    SXM             ; To eliminate optimization problem

        nop
        stl     A,*SP(0)        ; store original lawCompInput: A --> SP(0)
        abs     A               ; Work with positive # only
        add     #ULAWBIAS,A     ; Add ulaw bias value (33)
        exp     A               ; calculate leading zeros, place in T
        nop
        st      T,*SP(1)        ; Store T in local variable exponent: T --> SP(1)
        norm    A               ; left justify A
        sfta    A,-16           ; shift into low acc
        ld      #24,B           ; delta of segment and T (1 less)
        sub     *SP(1),B        ; calculate segment (1 less)
        sfta    B,4             ; adjust segment to proper bits
        add     A,-10,B         ; concatenate seg and quantization bits
        xor     #ULAWMASK,B     ; invert PCM result
        bitf    *SP(0),8000h    ; check sign bit of original lawCompInput
        nop
        xc      2,NTC           ; if input negative, skip next statement
        add     #PCMSIGN,B      ; if input positive, negate B
        nop
        ld      B,A             ; Return the companded value in A


        frame   #2
        nop
        popm    AR7
        popm    AR6
        popm    AR1
        popm    ST1
        popm    ST0
        RETCMD





*--------------------------------------------------------------------------*
*   A-LAW COMPRESSION

*    Input:
*
*      lawCompInput (14 bit 2-s complement value)
*
*    Returns:
*
*      a-law compressed value ready to transmission (8-bits)
*
*    Note:
*
*      Every other bit is inverted (for transmission).
*
*--------------------------------------------------------------------------*
_OtxDspAlawCompress
        nop
        pshm    ST0             ; *SP(6) = ST0
        pshm    ST1             ; *SP(5) = ST1
        pshm    AR1             ; *SP(4) = AR1
        pshm    AR6             ; *SP(3) = AR6
        pshm    AR7             ; *SP(2) = AR7
        nop
        frame   #-2             ;---- local variables -----
                                ; *SP(1) = exponent
                                ; *SP(0) = original lawCompInput saved here
                                ;---- arguments -----------
                                ; (A)    = lawCompInput: 16-bit value to be companded
                                ; *SP(7) = return address
        ssbx    SXM             ; To eliminate optimization problem
        nop
        stl     A,*SP(0)        ; store original lawCompInput: A --> SP(0)
        abs     A               ; Work with positive # only
        sfta    A,-1            ; Divide by 2. Our internal representation is 14 bits, but A-law can only handle 13 bits. So make word a 13 bit word (1 sign bit and 12 magnitude bits); Added by HJ
        ld      #32,B           ; load 128 into B
        sub     A,B             ; subtract A from B
        bc      lead1,BLEQ      ; branch if leading 1 present
                                ; this first situation is the exception to the pattern in alaw
        ld      A,-1,B          ; load quantization into B
        b       pcm             ; branch to masking step
lead1   exp     A               ; calculate leading zeros, place in T
        nop
        st      T,*SP(1)        ; Store T in local variable exponent: T --> SP(1)
        norm    A               ; left justify A
        sfta    A,-16           ; shift into low acc
        ld      #25,B           ; delta of segment and T (-1)
        sub     *SP(1),B        ; calculate segment (-1)
        sfta    B,4             ; adjust segment to proper bits
        add     A,-10,B         ; concatenate seg and quantization bits
pcm     xor     #ALAWMASK,B     ; invert PCM result
        bitf    *SP(0),8000h    ; check sign bit of original lawCompInput
        nop
        xc      2,NTC           ; if input negative, skip next statement
        add     #PCMSIGN,B      ; if input positive, negate B
        nop
        ld      B,A             ; Return the companded value in A

        frame   #2
        nop
        popm    AR7
        popm    AR6
        popm    AR1
        popm    ST1
        popm    ST0
        RETCMD



